<?php

declare(strict_types=1);

namespace mirzaev\vk\api\methods;

// Файлы проекта
use mirzaev\vk\robots\robot,
    mirzaev\vk\robots\group;

// Встроенные библиотеки
use stdClass,
    Exception;

/**
 * Режимы отправки сообщений
 */
enum message_send
{
    /** Обычная отправка */
    case simple;

    /** Проверка отправки */
    case check;
}

/**
 * Режимы генерации идентификатора сессии доставки сообщения
 */
enum random_id
{
    /** Генерация: time() (нельзя отправить более чем 1 сообщение в секунду) */
    case date;

    /** Генерация: rand() */
    case random;

    /** Генерация: random_int(0, PHP_INT_MAX) */
    case crypto;
}

/**
 * Сообщение
 *
 * @method public static function put(RobotAbstract $from, int $to, string $message, int $mode = 2) Отправить сообщение
 *
 * @see https://vk.com/dev/messages.send
 * @see https://vk.com/dev/messages.getById
 *
 * @package mirzaev\vk\api
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 *
 * @todo Доработать строгий режим отправки: проверку сообщения в беседе (не имеет ID сообщений)
 */
final class messages extends method
{
    /**
     * @var message_send $send_mode Режим отправки сообщений
     */
    protected message_send $send_mode = message_send::simple;

    /**
     * @var random_id $generate_mode Режим генерации идентификатора сессии доставки сообщения
     */
    protected random_id $generate_mode = random_id::crypto;

    /**
     * @var ?int $lat Географическая ширина
     */
    protected ?int $lat = null;

    /**
     * @var ?int $long Географическая долгота
     */
    protected ?int $long = null;

    /**
     * @var ?int $reply_to Идентификатор сообщения, на которое требуется ответить
     */
    protected ?int $reply_to = null;

    /**
     * @var ?array $forward_messages Идентификаторы пересылаемых сообщений
     */
    protected ?array $forward_messages = null;

    /**
     * @var ?string $sticker_id Идентификатор стикера
     */
    protected ?string $sticker_id = null;

    /**
     * @var ?string $payload Полезная нагрузка
     */
    protected ?string $payload = null;

    /**
     * @var bool $dont_parse_links Не создавать представление ссылки в сообщении?
     */
    protected bool $dont_parse_links = false;

    /**
     * @var bool $disable_mentions Отключить уведомление об упоминании в сообщении?
     */
    protected bool $disable_mentions = false;

    /**
     * @var ?string $intent Интент
     */
    protected ?string $intent = null;

    /**
     * @var ?string $subscribe_id Число, которое будет использоваться для работы с интентами
     */
    protected ?int $subscribe_id = null;

    /**
     * Конструктор
     *
     * @param robot $robot Робот
     * @param int|string|array|null $receiver Получатель
     * @param string|null $text Текст
     */
    public function __construct(
        protected robot $robot,
        int|string|array|null $receiver = null,
        protected string|null $text = null
    ) {
        if (isset($this->text, $receiver)) {
            // Быстрая отправка

            $this->send($receiver);
        }
    }

    /**
     * Записать текст
     *
     * @param string $text Текст
     */
    public function text(string $text): self
    {
        // Записать текст
        if (!isset($this->text)) {
            $this->text = $text;
        } else {
            $this->text .= $text;
        }

        return $this;
    }

    /**
     * Записать сообщения для пересылки
     *
     * @param $ids Идентификаторы сообщений
     */
    public function forward(...$ids): self
    {
        // Запись
        $this->forward = array_merge($this->forward ?? [], $ids);

        return $this;
    }

    /**
     * Записать сообщение для ответа
     *
     * @param $id Идентификатор сообщения
     */
    public function reply(int $id): self
    {
        // Запись
        $this->reply = $id;

        return $this;
    }

    /**
     * Отправить сообщение
     *
     * @see https://vk.com/dev/messages.send
     *
     * @param int|string|null $receiver Получатель
     * @param ?string $message Сообщение
     * @param ?int $lat Географическая ширина
     * @param ?int $long Географическая долгота
     * @param ?array $attachments Вложения
     * @param ?int $reply_to Идентификатор сообщения, на которое требуется ответить
     * @param ?array $forward_messages Идентификаторы пересылаемых сообщений
     * @param ?forward $forward Пересылаемые сообщения (в другой чат)
     * @param ?string $sticker_id Идентификатор стикера
     * @param ?keyboard $keyboard Инстанция клавиатуры
     * @param ?template $template Инстанция шаблона сообщения
     * @param ?string $payload Полезная нагрузка
     * @param bool $dont_parse_links Не создавать представление ссылки в сообщении?
     * @param bool $disable_mentions Отключить уведомление об упоминании в сообщении?
     * @param ?string $intent Интент
     * @param ?int $subscribe_id Число, которое будет использоваться для работы с интентами
     * @param ?int $random_id Идентификатор сессии доставки сообщения (защита от повторных отправок)
     *
     * @return int|array Идентификатор успешно отправленного сообщения или ответ сервера (подразумевается ошибка)
     *
     * @todo
     * 1. Написать обработчик ошибок возвращаемых ВКонтакте
     * 2. Добавить параметр forward (не путать с forward_messages)
     * 3. Добавить клавиатуру
     * 4. Добавить шаблоны сообщений
     * 5. Добавить content_source
     */
    public function send(
        int|string|null $receiver,
        ?string $message = null,
        ?int $lat = null,
        ?int $long = null,
        ?array $attachments = null,
        ?int $reply_to = null,
        ?array $forward_messages = null,
        // ?forward $forward = null,
        ?string $sticker_id = null,
        // ?keyboard $keyboard = null,
        // ?template $template = null,
        ?string $payload = null,
        bool $dont_parse_links = false,
        bool $disable_mentions = false,
        ?string $intent = null,
        ?int $subscribe_id = null,
        ?int $random_id = null,
    ): int|array {
        // Реинициализация настроек
        $this->robot->api->reinit();

        // Инициализация получателя
        if (is_int($receiver)) ($id = $receiver - 2000000000) > 0 ? $this->robot->api['peer_id'] = $receiver : $this->robot->api['chat_id'] = $id;
        else if (is_array($receiver)) $this->robot->api['peer_ids'] = implode(',', $receiver);
        else if (is_string($receiver)) $this->robot->api['domain'] = $receiver;

        // Инициализация идентификатора сессии доставки сообщения (защита от повторных отправок)
        $this->robot->api['random_id'] = $random_id ?? match ($this->generate_mode) {
            random_id::date => time(),
            random_id::random => rand(),
            random_id::crypto => random_int(0, PHP_INT_MAX),
            default => random_int(0, PHP_INT_MAX)
        };

        // Инициализация текста в настройках API
        if (isset($message)) $this->robot->api['message'] = $message;
        else if (isset($this->text)) $this->robot->api['message'] = $this->text;

        // Инициализация широты
        if (isset($lat)) $this->robot->api['lat'] = $lat;
        else if (isset($this->lat)) $this->robot->api['lat'] = $this->lat;

        // Инициализация долготы
        if (isset($long)) $this->robot->api['long'] = $long;
        else if (isset($this->long)) $this->robot->api['long'] = $this->tlongext;

        // Инициализация вложений
        if (isset($attachments)) $this->robot->api['attachment'] = implode(',', $attachments);
        else if (isset($this->robot->api->data) && $this->robot->api->__get('data') !== []) $this->robot->api['attachment'] = implode(',', $this->robot->api->__get('data'));

        // Инициализация сообщения, на которое требуется ответить
        if (isset($reply_to)) $this->robot->api['reply_to'] = $reply_to;
        else if (isset($this->reply)) $this->robot->api['reply_to'] = $this->reply;

        // Инициализация пересылаемых сообщений
        if (isset($forward_messages)) $this->robot->api['forward_messages'] = implode(',', $forward_messages);
        else if (isset($this->forward_messages)) $this->robot->api['forward_messages'] = implode(',', $this->forward_messages);

        // Инициализация стикера
        if (isset($sticker_id)) $this->robot->api['sticker_id'] = $sticker_id;
        else if (isset($this->sticker_id)) $this->robot->api['sticker_id'] = $this->sticker_id;

        // Инициализация полезной нагрузки
        if (isset($payload)) $this->robot->api['payload'] = $payload;
        else if (isset($this->payload)) $this->robot->api['payload'] = $this->payload;

        // Инициализация пользовательского соглашения
        // $this->robot->api['content_source'] = $this->robot->content_source;

        // Инициализация "не создавать представление ссылки в сообщении?"
        if ($dont_parse_links) $this->robot->api['dont_parse_links'] = 1;
        else if ($this->dont_parse_links) $this->robot->api['dont_parse_links'] = 1;

        // Инициализация "отключить уведомление об упоминании в сообщении?"
        if ($disable_mentions) $this->robot->api['disable_mentions'] = 1;
        else if ($this->disable_mentions) $this->robot->api['disable_mentions'] = 1;

        // Инициализация интентов
        if (isset($intent)) $this->robot->api['intent'] = $intent;
        else if (isset($this->intent)) $this->robot->api['intent'] = $this->intent;

        // Инициализация числа, которое будет использоваться для работы с интентами
        if (isset($subscribe_id)) $this->robot->api['subscribe_id'] = $subscribe_id;
        else if (isset($this->subscribe_id)) $this->robot->api['subscribe_id'] = $this->subscribe_id;

        // Проверка сформированного сообщения
        if (!$this->robot->api->offsetExists('message') && !$this->robot->api->offsetExists('attachment')) throw new Exception('Сообщение должно содержать текст, либо вложение');

        // Запрос
        $request = json_decode($this->robot->browser->request('POST', 'messages.send', ['form_params' => $this->robot->api->settings])->getBody()->getContents());

        // Если в ответе ошибка
        if (isset($request->error)) {
            throw new Exception('ВКонтакте: ' . $request->error->error_msg, $request->error->error_code);
        }

        if ($this->send_mode === message_send::check) {
            // Надёжная доставка сообщения

            if (!empty($request["response"])) {
                // Ответ получен

                // Реиницилазиция
                $this->robot->api->reinit();

                // Запрашиваемые сообщения
                $this->robot->api['message_ids'] = $request["response"];

                if ($this->robot->browser->request('POST', 'messages.getById', ['form_params' => $this->robot->api->settings])['response']['count'] === 0) {
                    // Сообщения не существует
                    //!!!!!!!!!!!!!!!!!!!!!!!!!!!
                    //!!!!!!!!!!!!!!!!!!!!!!!!!!
                    //!!!!!!!!!!!!!!!!!!!!!!!!!

                    // Повторная отправка
                    $this->send($receiver);
                }
            } else {
            }
        }

        return $request->response;
    }

    /**
     * Удалить сообщение
     *
     * @see https://vk.com/dev/messages.delete
     *
     * @param int|string|array|null $messages Получатель (message_ids + cmids)
     * @param int|string|null $peer_id Идентификатор беседы
     * @param bool $spam Пометить как спам?
     * @param bool $delete_for_all Удалить для всех?
     *
     * @return int|array Идентификатор успешно отправленного сообщения или ответ сервера (подразумевается ошибка)
     *
     * @todo Написать обработчик ошибок возвращаемых ВКонтакте
     */
    public function delete(string|int|array|null $messages = null, int|string|null $peer_id = null, bool $spam = false, bool $delete_for_all = false): stdClass
    {
        // Реиницилазиция настроек
        $this->robot->api->reinit();

        if (isset($peer_id)) {
            // Получен идентификатор беседы

            // Инициализация идентификатора беседы
            $this->robot->api['peer_id'] = $peer_id;

            // Инициализация идентификаторов сообщений
            $this->robot->api['cmids'] = $messages;

            // Инициализация: "удалить для всех?"
            $this->robot->api['delete_for_all'] = 1;
        } else {
            // Не получен идентификатор беседы

            // Инициализация идентификаторов сообщений
            $this->robot->api['message_ids'] = $messages;

            // Инициализация: "удалить для всех?"
            if ($delete_for_all) $this->robot->api['delete_for_all'] = 1;
        }

        // Инициализация: "сообщить о спаме?"
        if ($spam) $this->robot->api['spam'] = $spam;

        // Запрос
        $request = json_decode($this->robot->browser->request('POST', 'messages.delete', ['form_params' => $this->robot->api->settings])->getBody()->getContents());

        // Проверка на наличие ошибок в ответе от ВКонтакте
        if (isset($request->error)) throw new Exception('ВКонтакте: ' . $request->error->error_msg, $request->error->error_code);


        return $request->response;
    }

    /**
     * Записать свойство
     *
     * @param string $name Название
     * @param mixed $value Значение
     *
     * @return void
     */
    public function __set(string $name, mixed $value): void
    {
        match ($name) {
            default => throw new Exception("Свойство $name не найдено", 404)
        };
    }

    /**
     * Прочитать свойство
     *
     * @param string $name Название
     *
     * @return mixed
     */
    public function __get(string $name): mixed
    {
        return match ($name) {
            'text' => $this->text ?? throw new Exception('Текст не инициализирован'),
            'forward' => empty($this->forward) ? throw new Exception('Сообщения для пересылки не инициализированы') : $this->forward,
            'reply' => isset($this->reply) ? $this->reply : throw new Exception('Сообщение для ответа не инициализировано'),
            default => throw new Exception("Свойство $name не найдено", 404)
        };
    }

    /**
     * Получить информацию о сообщении по id
     *
     * @param string $message_ids Идентификатор сообщения
     *
     * @return object Информация о сообщении
     */

    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // доделать

    public function getById(int $message_ids): object
    {
        // Реиницилазиция
        $this->robot->api->reinit();

        // Идентификаторы сообщений. Максимум 100 идентификаторов
        $this->robot->api['message_ids'] = $message_ids;


        if ($this->robot instanceof group) {
            // Если это группа

            // Идентификатор группы
            isset($this->robot->id) ? $this->robot->api['group_id'] = $this->robot->id : throw new Exception('Необходимо указать идентификатор сообщества');
        }

        // Запрос
        $request = json_decode(($this->robot->browser->request('POST', 'messages.getById', ['form_params' => $this->robot->api->settings])->getBody()->getContents()));

        // Если в ответе ошибка
        if (isset($request->error)) {
            throw new Exception('Вконтакте: ' . $request->error->error_msg, $request->error->error_code);
        }

        return $request->response;
    }
}
